' VB.NET port of the original Gpc C# wrapper written by Stefan Menne
'   V1.0
'   Tadej Basa
'   tadej.basa@gmail.com
'   8.12.2004
'
'   For more details: http://www.cs.man.ac.uk/aig/staff/alan/software/
'   
' Original C# wrapper by:
'   Stefan Menne
'   stefan.men@web.de
'   10.3.2004

Imports System
Imports System.Collections
Imports System.Windows.Forms
Imports System.Drawing
Imports System.Drawing.Drawing2D
Imports System.Runtime.InteropServices

Namespace GpcWrapper
    Public Enum GpcOperation
        _Difference = 0
        _Intersection = 1
        _XOr = 2
        _Union = 3
    End Enum

    Public Structure Vertex
        Public X As Double
        Public Y As Double

        Public Sub New(ByVal x As Double, ByVal y As Double)
            x = x
            y = y
        End Sub

        Public Overrides Function ToString() As String
            Return "(" + X.ToString() + "," + Y.ToString() + ")"
        End Function
    End Structure

    Public Class VertexList
        Public NofVertices As Integer
        Public Vertex() As Vertex

        Public Sub New()
        End Sub

        Public Sub New(ByVal p() As PointF)
            NofVertices = p.Length
            ReDim Vertex(NofVertices - 1)
            Dim i As Integer
            For i = 0 To p.Length - 1
                Vertex(i).X = p(i).X
                Vertex(i).Y = p(i).Y
            Next
        End Sub

        Public Function ToGraphicsPath() As GraphicsPath
            Dim graphicsPath As GraphicsPath = New GraphicsPath
            graphicsPath.AddLines(ToPoints())
            Return graphicsPath
        End Function

        Public Function ToPoints() As PointF()

            Dim vertexArray(NofVertices - 1) As PointF
            Dim i As Integer
            For i = 0 To NofVertices - 1
                vertexArray(i).X = Vertex(i).X
                vertexArray(i).Y = Vertex(i).Y
            Next
            Return vertexArray
        End Function

        Public Function TristripToGraphicsPath() As GraphicsPath
            Dim graphicsPath As GraphicsPath = New GraphicsPath

            Dim i As Integer
            For i = 0 To NofVertices - 2 - 1
                graphicsPath.AddPolygon(New PointF(2) {New PointF(CType(Vertex(i).X, Single), CType(Vertex(i).Y, Single)), _
                                                       New PointF(CType(Vertex(i + 1).X, Single), CType(Vertex(i + 1).Y, Single)), _
                                                       New PointF(CType(Vertex(i + 2).X, Single), CType(Vertex(i + 2).Y, Single))})
            Next i
            Return graphicsPath
        End Function

        Public Overrides Function ToString() As String
            Dim s As String = "Polygon with " + NofVertices + " vertices: "

            Dim i As Integer
            For i = 0 To NofVertices - 1
                s += Vertex(i).ToString()
                If (i <> NofVertices - 1) Then s += ","
            Next
            Return s
        End Function
    End Class

    Public Class Polygon
        Public NofContours As Integer
        Public ContourIsHole() As Boolean
        Public Contour() As VertexList

        Public Sub New()
        End Sub

        'path should contain only polylines ( use Flatten )
        'furthermore the constructor assumes that all Subpathes of path except the first one are holes
        Public Sub New(ByVal path As GraphicsPath)
            NofContours = 0
            Dim b As Byte
            For Each b In path.PathTypes
                If (b And (CType(PathPointType.CloseSubpath, Byte))) <> 0 Then
                    NofContours = NofContours + 1
                End If
            Next

            ReDim ContourIsHole(NofContours - 1)
            ReDim Me.Contour(NofContours - 1)

            Dim i As Integer
            For i = 0 To NofContours - 1
                ContourIsHole(i) = (i = 0)
            Next

            Dim contourNr As Integer = 0
            Dim cont As ArrayList = New ArrayList
            For i = 0 To path.PathPoints.Length - 1
                cont.Add(path.PathPoints(i))
                If (path.PathTypes(i) And (CType(PathPointType.CloseSubpath, Byte))) <> 0 Then
                    Dim pointArray() As PointF = cont.ToArray(GetType(PointF))
                    Dim vl As VertexList = New VertexList(pointArray)
                    Contour(contourNr) = vl
                    contourNr += 1
                    cont.Clear()
                End If
            Next
        End Sub

        Public Shared Function FromFile(ByVal filename As String, ByVal readHoleFlags As Boolean) As Polygon
            Return GpcWrapper.ReadPolygon(filename, readHoleFlags)
        End Function

        Public Sub AddContour(ByVal c As VertexList, ByVal contIsHole As Boolean)
            Dim hole(NofContours) As Boolean
            Dim cont(NofContours) As VertexList

            Dim i As Integer
            For i = 0 To NofContours - 1
                hole(i) = Me.ContourIsHole(i)
                cont(i) = Me.Contour(i)
            Next
            hole(NofContours) = contIsHole
            cont(NofContours) = c
            NofContours += 1

            ContourIsHole = hole
            Contour = cont
        End Sub

        Public Function ToGraphicsPath() As GraphicsPath
            Dim path As GraphicsPath = New GraphicsPath

            Dim i As Integer
            For i = 0 To NofContours - 1
                Dim points() As PointF = Contour(i).ToPoints()
                If (ContourIsHole(i)) Then Array.Reverse(points)
                path.AddPolygon(points)
            Next
            Return path
        End Function

        Public Overrides Function ToString() As String
            Dim s As String = "Polygon with " + NofContours.ToString() + " contours." + "\r\n"
            Dim i As Integer
            For i = 0 To NofContours - 1
                If (ContourIsHole(i)) Then
                    s += "Hole: "
                Else
                    s += "Contour: "
                End If
                s += Contour(i).ToString()
            Next
            Return s
        End Function

        Public Function ClipToTristrip(ByVal operation As GpcOperation, ByVal polygon As Polygon) As Tristrip
            Return GpcWrapper.ClipToTristrip(operation, Me, polygon)
        End Function

        Public Function Clip(ByVal operation As GpcOperation, ByVal polygon As Polygon) As Polygon
            Return GpcWrapper.Clip(operation, Me, polygon)
        End Function

        Public Function ToTristrip() As Tristrip
            Return GpcWrapper.PolygonToTristrip(Me)
        End Function

        Public Sub Save(ByVal filename As String, ByVal writeHoleFlags As Boolean)
            GpcWrapper.SavePolygon(filename, writeHoleFlags, Me)
        End Sub
    End Class

    Public Class Tristrip
        Public NofStrips As Integer
        Public Strip() As VertexList
    End Class

    Public Class GpcWrapper
        Public Shared Function PolygonToTristrip(ByVal poly As Polygon) As Tristrip

            Dim gpc_str As gpc_tristrip = New gpc_tristrip
            Dim gpc_pol As gpc_polygon = GpcWrapper.PolygonTo_gpc_polygon(poly)
            gpc_polygon_to_tristrip(gpc_pol, gpc_str)
            Dim tristrip As Tristrip = GpcWrapper.gpc_strip_ToTristrip(gpc_str)

            GpcWrapper.Free_gpc_polygon(gpc_pol)
            GpcWrapper.gpc_free_tristrip(gpc_str)

            Return tristrip
        End Function


        Public Shared Function ClipToTristrip(ByVal operation As GpcOperation, ByVal subject_polygon As Polygon, ByVal clip_polygon As Polygon) As Tristrip
            Dim gpc_strip As gpc_tristrip = New gpc_tristrip
            Dim gpc_subject_polygon As gpc_polygon = GpcWrapper.PolygonTo_gpc_polygon(subject_polygon)
            Dim gpc_clip_polygon As gpc_polygon = GpcWrapper.PolygonTo_gpc_polygon(clip_polygon)

            gpc_tristrip_clip(operation, gpc_subject_polygon, gpc_clip_polygon, gpc_strip)
            Dim tristrip As Tristrip = GpcWrapper.gpc_strip_ToTristrip(gpc_strip)

            GpcWrapper.Free_gpc_polygon(gpc_subject_polygon)
            GpcWrapper.Free_gpc_polygon(gpc_clip_polygon)
            GpcWrapper.gpc_free_tristrip(gpc_strip)

            Return tristrip
        End Function


        Public Shared Function Clip(ByVal operation As GpcOperation, ByVal subject_polygon As Polygon, ByVal clip_polygon As Polygon) As Polygon
            Dim gpc_polygon As gpc_polygon = New gpc_polygon
            Dim gpc_subject_polygon As gpc_polygon = GpcWrapper.PolygonTo_gpc_polygon(subject_polygon)
            Dim gpc_clip_polygon As gpc_polygon = GpcWrapper.PolygonTo_gpc_polygon(clip_polygon)

            gpc_polygon_clip(operation, gpc_subject_polygon, gpc_clip_polygon, gpc_polygon)
            Dim polygon As Polygon = GpcWrapper.gpc_polygon_ToPolygon(gpc_polygon)

            GpcWrapper.Free_gpc_polygon(gpc_subject_polygon)
            GpcWrapper.Free_gpc_polygon(gpc_clip_polygon)
            GpcWrapper.gpc_free_polygon(gpc_polygon)

            Return polygon
        End Function



        Public Shared Sub SavePolygon(ByVal filename As String, ByVal writeHoleFlags As Boolean, ByVal polygon As Polygon)
            Dim gpc_poly As gpc_polygon = GpcWrapper.PolygonTo_gpc_polygon(polygon)

            Dim fp As IntPtr = fopen(filename, "w")
            gpc_write_polygon(fp, IIf(writeHoleFlags, 1, 0), gpc_poly)
            fclose(fp)

            GpcWrapper.Free_gpc_polygon(gpc_poly)
        End Sub



        Public Shared Function ReadPolygon(ByVal filename As String, ByVal readHoleFlags As Boolean) As Polygon
            Dim gpc_polygon As gpc_polygon = New gpc_polygon

            Dim fp As IntPtr = fopen(filename, "r")
            gpc_read_polygon(fp, IIf(readHoleFlags, 1, 0), gpc_polygon)
            Dim polygon As Polygon = gpc_polygon_ToPolygon(gpc_polygon)
            gpc_free_polygon(gpc_polygon)
            fclose(fp)

            Return polygon
        End Function


        Private Shared Function PolygonTo_gpc_polygon(ByVal polygon As Polygon) As gpc_polygon
            Dim gpc_pol As gpc_polygon = New gpc_polygon
            gpc_pol.num_contours = polygon.NofContours

            Dim hole(polygon.NofContours - 1) As Integer
            Dim i As Integer
            For i = 0 To polygon.NofContours - 1
                hole(i) = IIf(polygon.ContourIsHole(i), 1, 0)
            Next
            gpc_pol.hole = Marshal.AllocCoTaskMem(polygon.NofContours * Marshal.SizeOf(hole(0)))
            Marshal.Copy(hole, 0, gpc_pol.hole, polygon.NofContours)

            gpc_pol.contour = Marshal.AllocCoTaskMem(polygon.NofContours * Marshal.SizeOf(New gpc_vertex_list))
            Dim ptr As IntPtr = gpc_pol.contour

            For i = 0 To polygon.NofContours - 1
                Dim gpc_vtx_list As gpc_vertex_list = New gpc_vertex_list
                gpc_vtx_list.num_vertices = polygon.Contour(i).NofVertices
                gpc_vtx_list.vertex = Marshal.AllocCoTaskMem(polygon.Contour(i).NofVertices * Marshal.SizeOf(New gpc_vertex))
                Dim ptr2 As IntPtr = gpc_vtx_list.vertex
                Dim j As Integer
                For j = 0 To polygon.Contour(i).NofVertices - 1
                    Dim gpc_vtx As gpc_vertex = New gpc_vertex
                    gpc_vtx.x = polygon.Contour(i).Vertex(j).X
                    gpc_vtx.y = polygon.Contour(i).Vertex(j).Y
                    Marshal.StructureToPtr(gpc_vtx, ptr2, False)
                    ptr2 = New IntPtr(ptr2.ToInt32 + Marshal.SizeOf(gpc_vtx))
                Next
                Marshal.StructureToPtr(gpc_vtx_list, ptr, False)
                ptr = New IntPtr(CType(ptr.ToInt32, Integer) + Marshal.SizeOf(gpc_vtx_list))
            Next

            Return gpc_pol
        End Function


        Private Shared Function gpc_polygon_ToPolygon(ByVal gpc_poly As gpc_polygon) As Polygon
            Dim polygon As Polygon = New Polygon

            If gpc_poly.num_contours = 0 Then Return polygon

            polygon.NofContours = gpc_poly.num_contours
            ReDim polygon.ContourIsHole(polygon.NofContours - 1)
            ReDim polygon.Contour(polygon.NofContours - 1)
            Dim holeShort(polygon.NofContours - 1) As Short

            Dim ptr As IntPtr = gpc_poly.hole
            Marshal.Copy(gpc_poly.hole, holeShort, 0, polygon.NofContours)

            Dim i As Integer
            For i = 0 To polygon.NofContours - 1
                polygon.ContourIsHole(i) = (holeShort(i) <> 0)
            Next

            ptr = gpc_poly.contour
            For i = 0 To polygon.NofContours - 1
                Dim gpc_vtx_list As gpc_vertex_list = CType(Marshal.PtrToStructure(ptr, GetType(gpc_vertex_list)), gpc_vertex_list)
                polygon.Contour(i) = New VertexList
                polygon.Contour(i).NofVertices = gpc_vtx_list.num_vertices
                ReDim polygon.Contour(i).Vertex(polygon.Contour(i).NofVertices - 1)
                Dim ptr2 As IntPtr = gpc_vtx_list.vertex
                Dim j As Integer
                For j = 0 To polygon.Contour(i).NofVertices - 1
                    Dim gpc_vtx As gpc_vertex = CType(Marshal.PtrToStructure(ptr2, GetType(gpc_vertex)), gpc_vertex)
                    polygon.Contour(i).Vertex(j).X = gpc_vtx.x
                    polygon.Contour(i).Vertex(j).Y = gpc_vtx.y
                    ptr2 = New IntPtr(CType(ptr2.ToInt32, Integer) + Marshal.SizeOf(gpc_vtx))
                Next
                ptr = New IntPtr(CType(ptr.ToInt32, Integer) + Marshal.SizeOf(gpc_vtx_list))
            Next
            Return polygon
        End Function

        Private Shared Function gpc_strip_ToTristrip(ByVal gpc_strip As gpc_tristrip) As Tristrip
            Dim tristrip As Tristrip = New Tristrip
            tristrip.NofStrips = gpc_strip.num_strips
            tristrip.Strip = New VertexList(tristrip.NofStrips) {}
            Dim ptr As IntPtr = gpc_strip.strip
            Dim i As Integer
            For i = 0 To tristrip.NofStrips - 1
                tristrip.Strip(i) = New VertexList
                Dim gpc_vtx_list As gpc_vertex_list = CType(Marshal.PtrToStructure(ptr, GetType(gpc_vertex_list)), gpc_vertex_list)
                tristrip.Strip(i).NofVertices = gpc_vtx_list.num_vertices
                tristrip.Strip(i).Vertex = New Vertex(tristrip.Strip(i).NofVertices) {}

                Dim ptr2 As IntPtr = gpc_vtx_list.vertex
                Dim j As Integer
                For j = 0 To tristrip.Strip(i).NofVertices - 1
                    Dim gpc_vtx As gpc_vertex = CType(Marshal.PtrToStructure(ptr2, GetType(gpc_vertex)), gpc_vertex)
                    tristrip.Strip(i).Vertex(j).X = gpc_vtx.x
                    tristrip.Strip(i).Vertex(j).Y = gpc_vtx.y
                    ptr2 = New IntPtr(CType(ptr2.ToInt32, Integer) + Marshal.SizeOf(gpc_vtx))
                Next
                ptr = New IntPtr(CType(ptr.ToInt32, Integer) + Marshal.SizeOf(gpc_vtx_list))
            Next

            Return tristrip
        End Function

        Private Shared Sub Free_gpc_polygon(ByVal gpc_pol As gpc_polygon)
            Marshal.FreeCoTaskMem(gpc_pol.hole)
            Dim ptr As IntPtr = gpc_pol.contour
            Dim i As Integer
            For i = 0 To gpc_pol.num_contours - 1
                Dim gpc_vtx_list As gpc_vertex_list = CType(Marshal.PtrToStructure(ptr, GetType(gpc_vertex_list)), gpc_vertex_list)
                Marshal.FreeCoTaskMem(gpc_vtx_list.vertex)
                ptr = New IntPtr(CType(ptr.ToInt32, Integer) + Marshal.SizeOf(gpc_vtx_list))
            Next
        End Sub

        Public Declare Auto Sub gpc_polygon_to_tristrip Lib "gpc.dll" (ByRef poly As gpc_polygon, _
                                                                       ByRef tripstrip As gpc_tristrip)

        Public Declare Auto Sub gpc_polygon_clip Lib "gpc.dll" (ByVal set_operation As GpcOperation, _
                                                                ByRef subject_polygon As gpc_polygon, _
                                                                ByRef clip_polygon As gpc_polygon, _
                                                                ByRef result_polygon As gpc_polygon)

        Public Declare Auto Sub gpc_tristrip_clip Lib "gpc.dll" (ByVal set_operation As GpcOperation, _
                                                                 ByRef subject_polygon As gpc_polygon, _
                                                                 ByRef clip_polygon As gpc_polygon, _
                                                                 ByRef result_tristrip As gpc_tristrip)

        Public Declare Auto Sub gpc_free_tristrip Lib "gpc.dll" (ByRef hwnd As gpc_tristrip)

        Public Declare Auto Sub gpc_free_polygon Lib "gpc.dll" (ByRef poly As gpc_polygon)

        Public Declare Auto Sub gpc_read_polygon Lib "gpc.dll" (ByVal fp As IntPtr, ByVal read_hole_flags As Integer, ByRef poly As gpc_polygon)

        Public Declare Auto Sub gpc_write_polygon Lib "gpc.dll" (ByVal fp As IntPtr, ByVal read_hole_flags As Integer, ByRef poly As gpc_polygon)

        Public Declare Auto Function fopen Lib "msvcr71.dll" (ByVal fn As String, ByVal mode As String) As IntPtr

        Public Declare Auto Sub fclose Lib "msvcr71.dll" (ByVal fp As IntPtr)

        Public Declare Auto Function fputc Lib "msvcr71.dll" (ByVal c As Integer, ByVal fp As IntPtr) As Integer


        Enum gpc_op
            GPC_DIFF = 0
            GPC_INT = 1
            GPC_XOR = 2
            GPC_UNION = 3
        End Enum

        'Point
        <StructLayout(LayoutKind.Sequential)> _
        Public Structure gpc_vertex
            Public x As Double
            Public y As Double
        End Structure

        'Point list
        <StructLayout(LayoutKind.Sequential)> _
        Public Structure gpc_vertex_list
            Public num_vertices As Integer
            Public vertex As IntPtr
        End Structure


        <StructLayout(LayoutKind.Sequential)> _
       Public Structure gpc_polygon
            Public num_contours As Integer
            Public hole As IntPtr
            Public contour As IntPtr
        End Structure

        <StructLayout(LayoutKind.Sequential)> _
        Public Structure gpc_tristrip
            Public num_strips As Integer
            Public strip As IntPtr
        End Structure
    End Class
End Namespace


